
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_INTERPOLATOR_HH_
#define _MOOF_INTERPOLATOR_HH_

#include <Moof/Log.hh>
#include <Moof/Math.hh>


namespace Mf {


namespace Interp {

	typedef enum
	{
		STOP		= 0,
		REPEAT		= 1,
		OSCILLATE	= 2
	} Mode;

} // namespace Interp


template <class T>
class Interpolator : public T
{
public:

	Interpolator(Scalar t = 1.0, Interp::Mode mode = Interp::STOP)
	{
		reset(t, mode);
	}

	void reset(Scalar t = 1.0, Interp::Mode mode = Interp::STOP)
	{
		mAlpha = 0.0;
		mScale = 1.0 / t;
		mMode = mode;
		mIsDone = false;
	}

	void update(Scalar t, Scalar dt)
	{
		if (!mIsDone)
		{
			mPrevState = T::getValue();
			mAlpha += dt * mScale;
			clamp();
			if (mPrevState == T::calculate(mAlpha)) mIsDone = true;
		}
	}

	typename T::Type getState(Scalar alpha) const
	{
		return cml::lerp(mPrevState, T::getValue(), alpha);
	}

	bool isDone() const
	{
		return mIsDone;
	}

private:

	void clamp()
	{
		if (mAlpha > 1.0)
		{
			switch (mMode)
			{
				case Interp::STOP:
					mAlpha = SCALAR(1.0);
					break;
				case Interp::REPEAT:
					mAlpha -= SCALAR(1.0);
					break;
				case Interp::OSCILLATE:
					mAlpha = SCALAR(2.0) - mAlpha;
					mScale = -mScale;
					break;
			}
		}
		else if (mAlpha < 0.0)
		{
			switch (mMode)
			{
				case Interp::STOP:
					mAlpha = SCALAR(0.0);
					break;
				case Interp::REPEAT:
					mAlpha += SCALAR(1.0);
					break;
				case Interp::OSCILLATE:
					mAlpha = -mAlpha;
					mScale = -mScale;
					break;
			}
		}
	}

	Scalar				mAlpha;
	Scalar				mScale;
	Interp::Mode		mMode;
	bool				mIsDone;

	typename T::Type	mPrevState;
};


template <typename T = Scalar>
class Linear
{
public:

	typedef T Type;

	void init(const Type& a, const Type& b)
	{
		mStart = a;
		mFinish = b;
	}

	const Type& calculate(Scalar alpha)
	{
		mState = cml::lerp(mStart, mFinish, alpha);
		return mState;
	}

	const Type& getValue() const
	{
		return mState;
	}

private:

	Type	mState;
	Type	mStart;
	Type	mFinish;
};


typedef Interpolator< Linear<Scalar> > Lerp;


} // namespace Mf

#endif // _MOOF_INTERPOLATOR_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

