
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_TEXTURE_HH_
#define _MOOF_TEXTURE_HH_

/**
 * @file Texture.hh
 * Image-loading and OpenGL texture loading.
 */

#include <boost/shared_ptr.hpp>

#include <Moof/Image.hh>
#include <Moof/OpenGL.hh>


namespace Mf {


class Texture;
typedef boost::shared_ptr<Texture> TextureP;


class Texture : public Image
{
public:

	/**
	 * Possible orientations for texture coordinates.
	 */

	typedef unsigned TileIndex;
	static const TileIndex NO_TILE = -1;

	typedef enum
	{
		NORMAL =			0,			///< Normal orientation.
		FLIP =				1,			///< Flip over a horizontal axis.
		REVERSE =			2,			///< Flip over a vertical axis.
		FLIP_AND_REVERSE =	3			///< Flip over both.
	} Orientation;


	static TextureP alloc(const std::string& name)
	{
		return TextureP(new Texture(name));
	}

	explicit Texture(const std::string& name);

	void bind() const;
	GLuint getObject() const;

	static void resetBind();

	void setMinFilter(GLuint filter);
	void setMagFilter(GLuint filter);
	void setWrapS(GLuint wrap);
	void setWrapT(GLuint wrap);


	/**
	 * Calculate texture coordinates for a tile at a certain index.  Tiles are
	 * indexed start with zero as the to-left tile and moving across, then down.
	 * @param index The tile index.
	 * @param coords An array of scalars where the texture coordinates will be
	 * stored after this call.  The first coordinate (u,v) will be in the first
	 * two places and so on until all four coordinates are stored, therefore
	 * requiring enough room for an array of eight scalars.  The winding of the
	 * coordinates is always counter-clockwise (the GL default).
	 * @return True if index is valid, false otherwise.
	 */

	bool getTileCoords(TileIndex index, Scalar coords[8]) const;


	/**
	 * This version let's you specify an orientation that will be reflected in
	 * the texture coordinates.  This allows you to easily map a texture
	 * backwards or upside-down.
	 * @param what The orientation; can be flip, reverse, or flip_and_reverse.
	 * @return True if index is valid, false otherwise.
	 */

	bool getTileCoords(TileIndex index, Scalar coords[8], Orientation what) const;


	static std::string getPath(const std::string& name);

private:

	class Impl;
	boost::shared_ptr<Impl> mImpl;
};


} // namespace Mf

#endif // _MOOF_TEXTURE_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

